"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _Room_instances, _a, _Room_CONTROL_CHANNEL, _Room__state, _Room__role, _Room_inviteId, _Room__pc, _Room_incomingMessageQueue, _Room_dChannel, _Room__creationTime, _Room_dChannelSend, _Room_applyContentTypeConfig, _Room_setRemoteDescription;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Room = void 0;
const uuid_1 = require("uuid");
const promises_1 = __importDefault(require("./promises"));
const notifier_1 = require("./notifier");
const constants_1 = require("./constants");
const logger_1 = __importDefault(require("./logger"));
class Room {
    constructor(connection, name, pin, nickname, creationTime) {
        _Room_instances.add(this);
        _Room__state.set(this, constants_1.RoomState.NEW);
        _Room__role.set(this, constants_1.ParticipantRole.PARTICIPANT);
        _Room_inviteId.set(this, void 0);
        this.notifier = new notifier_1.Notifier();
        _Room__pc.set(this, void 0);
        _Room_incomingMessageQueue.set(this, {});
        _Room_dChannel.set(this, void 0);
        _Room__creationTime.set(this, void 0);
        this.connection = connection;
        this._id = name;
        this._name = name;
        this._nickname = nickname;
        this._pin = pin;
        this._nickname = nickname;
        __classPrivateFieldSet(this, _Room__creationTime, creationTime, "f");
        this.logger = new logger_1.default();
        this._uid = (0, uuid_1.v4)();
        this._crutch = {
            settingRemoteOffer: false,
            settingRemoteAnswer: false,
            tid: ""
        };
    }
    ;
    ;
    ;
    //TODO refactor types
    processEvent(e) {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.info("<==", e);
            if (e.type === constants_1.RoomEvent.REMOTE_SDP) {
                if (__classPrivateFieldGet(this, _Room__state, "f") !== constants_1.RoomState.FAILED && __classPrivateFieldGet(this, _Room__state, "f") !== constants_1.RoomState.DISPOSED && __classPrivateFieldGet(this, _Room__pc, "f").signalingState !== "closed") {
                    try {
                        const remoteSdp = e;
                        switch (remoteSdp.info.type) {
                            case constants_1.RemoteSdpType.OFFER:
                                if (this._crutch.settingRemoteAnswer) {
                                    this.logger.debug(new Error("Clashing with the answer!"));
                                    function sleep(millis) {
                                        return new Promise(resolve => setTimeout(resolve, millis));
                                    }
                                    while (this._crutch.settingRemoteAnswer) {
                                        yield sleep(100);
                                        this.logger.debug("Waiting for answer " + this._crutch.settingRemoteAnswer);
                                    }
                                }
                                this._crutch.tid = remoteSdp.info.tid;
                                this.logger.debug("setTid " + this._crutch.tid);
                                this._crutch.settingRemoteOffer = true;
                                if (__classPrivateFieldGet(this, _Room__pc, "f").signalingState === "have-local-offer") {
                                    yield __classPrivateFieldGet(this, _Room_instances, "m", _Room_setRemoteDescription).call(this, {
                                        info: {
                                            sdp: "",
                                            type: constants_1.RemoteSdpType.ROLLBACK,
                                            tid: ""
                                        }
                                    });
                                }
                                yield __classPrivateFieldGet(this, _Room_instances, "m", _Room_setRemoteDescription).call(this, remoteSdp);
                                const answer = yield __classPrivateFieldGet(this, _Room__pc, "f").createAnswer();
                                answer.sdp = answer.sdp.replace(/a=sendrecv/g, "a=sendonly");
                                yield __classPrivateFieldGet(this, _Room__pc, "f").setLocalDescription(answer);
                                if (__classPrivateFieldGet(this, _Room__pc, "f").connectionState !== "closed") {
                                    this.connection.send(constants_1.InternalApi.UPDATE_ROOM_STATE, {
                                        id: this._id,
                                        pin: this._pin,
                                        sdp: __classPrivateFieldGet(this, _Room__pc, "f").localDescription.sdp,
                                        tid: this._crutch.tid,
                                        sdpType: constants_1.RemoteSdpType.ANSWER
                                    });
                                }
                                this._crutch.settingRemoteOffer = false;
                                break;
                            case constants_1.RemoteSdpType.ANSWER:
                                this._crutch.settingRemoteAnswer = true;
                                yield __classPrivateFieldGet(this, _Room_instances, "m", _Room_setRemoteDescription).call(this, remoteSdp);
                                this._crutch.settingRemoteAnswer = false;
                                break;
                        }
                    }
                    catch (error) {
                        this.logger.error("Failed to process remote sdp", error);
                    }
                }
            }
            else if (e.type === constants_1.RoomEvent.ROLE_ASSIGNED) {
                const roleAssigned = e;
                if (this._nickname === roleAssigned.name) {
                    __classPrivateFieldSet(this, _Room__role, roleAssigned.role, "f");
                }
                this.notifier.notify(constants_1.RoomEvent.ROLE_ASSIGNED, roleAssigned);
            }
            else if (e.type === constants_1.RoomEvent.CREATED) {
                const createdRoom = e;
                __classPrivateFieldSet(this, _Room_inviteId, createdRoom.inviteId, "f");
                if (!promises_1.default.resolve(createdRoom.internalMessageId, e)) {
                    this.notifier.notify(constants_1.RoomEvent.CREATED, createdRoom);
                }
            }
            else if (e.type === constants_1.RoomEvent.OPERATION_FAILED) {
                const operationFailed = e;
                if (promises_1.default.promised(operationFailed.internalMessageId)) {
                    promises_1.default.reject(operationFailed.internalMessageId, operationFailed);
                }
                else if (operationFailed.operation === constants_1.Operations.ROOM_JOIN && operationFailed.error === constants_1.RoomError.ROOM_DESTROYED) {
                    this.notifier.notify(constants_1.RoomEvent.ENDED);
                }
                else if (operationFailed.operation === constants_1.Operations.ROOM_JOIN && operationFailed.error === constants_1.RoomError.AUTHORIZATION_FAILED) {
                    const evictedEvent = Object.assign(Object.assign({}, operationFailed), { name: this._nickname });
                    this.notifier.notify(constants_1.RoomEvent.EVICTED, evictedEvent);
                }
                else if (operationFailed.operation === constants_1.Operations.ROOM_JOIN && operationFailed.error === constants_1.RoomError.CANCEL_JOIN_ROOM) {
                    const leftRoomEvent = Object.assign(Object.assign({}, operationFailed), { name: this._nickname });
                    this.notifier.notify(constants_1.RoomEvent.LEFT, leftRoomEvent);
                }
                else {
                    this.notifier.notify(constants_1.RoomEvent.OPERATION_FAILED, operationFailed);
                }
            }
            else if (e.type === constants_1.RoomEvent.JOINED) {
                const joinedRoom = e;
                if (!promises_1.default.resolve(joinedRoom.internalMessageId, joinedRoom)) {
                    this.notifier.notify(constants_1.RoomEvent.JOINED, joinedRoom);
                }
            }
            else if (e.type === constants_1.RoomEvent.LEFT) {
                const leftRoom = e;
                promises_1.default.resolve(leftRoom.internalMessageId, leftRoom);
                this.notifier.notify(constants_1.RoomEvent.LEFT, leftRoom);
            }
            else if (e.type === constants_1.RoomEvent.PARTICIPANT_RENAMED) {
                const participantRenamed = e;
                if (this._nickname === participantRenamed.previousName) {
                    this._nickname = participantRenamed.updatedName;
                }
                if (!promises_1.default.promised(participantRenamed.internalMessageId)) {
                    this.notifier.notify(constants_1.RoomEvent.PARTICIPANT_RENAMED, participantRenamed);
                }
            }
            else if (e.type === constants_1.RoomEvent.ADD_TRACKS) {
                const addTracks = e;
                if (!promises_1.default.resolve(addTracks.internalMessageId, addTracks)) {
                    this.notifier.notify(constants_1.RoomEvent.ADD_TRACKS, addTracks);
                }
            }
            else if (e.type === constants_1.RoomEvent.REMOVE_TRACKS) {
                const removeTracks = e;
                if (!promises_1.default.resolve(removeTracks.internalMessageId, removeTracks)) {
                    this.notifier.notify(constants_1.RoomEvent.REMOVE_TRACKS, removeTracks);
                }
            }
            else if (e.type === constants_1.RoomEvent.WAITING_ROOM_UPDATE) {
                const waitingRoomUpdate = e;
                if (promises_1.default.promised(e.internalMessageId)) {
                    promises_1.default.resolve(e.internalMessageId, waitingRoomUpdate.enabled);
                }
                else {
                    this.notifier.notify(constants_1.RoomEvent.WAITING_ROOM_UPDATE, waitingRoomUpdate);
                }
            }
            else if (e.type === constants_1.SfuEvent.ACK && promises_1.default.promised(e.internalMessageId)) {
                promises_1.default.resolve(e.internalMessageId);
            }
            else {
                //check this is a room event
                if (Object.values(constants_1.RoomEvent).includes(e.type)) {
                    this.notifier.notify(e.type, e);
                }
            }
        });
    }
    //TODO(naz): safe guard based on state
    join(pc, nickname, config) {
        const self = this;
        __classPrivateFieldSet(this, _Room__pc, pc, "f");
        __classPrivateFieldGet(this, _Room__pc, "f").addEventListener("signalingstatechange", () => {
            this.logger.debug("Connection " + self._uid + " " + __classPrivateFieldGet(self, _Room__pc, "f").signalingState);
        });
        if (nickname) {
            this._nickname = nickname;
        }
        __classPrivateFieldSet(this, _Room_dChannel, __classPrivateFieldGet(this, _Room__pc, "f").createDataChannel(__classPrivateFieldGet(Room, _a, "f", _Room_CONTROL_CHANNEL)), "f");
        __classPrivateFieldGet(this, _Room_dChannel, "f").onmessage = (msg) => {
            this.logger.info("dchannel ", "<==", msg);
            const message = JSON.parse(msg.data);
            if (message.type === constants_1.RoomEvent.MESSAGE) {
                const msg = message;
                if (msg.message.message.indexOf("\"payload\":") !== -1) {
                    try {
                        let innerMessage = JSON.parse(msg.message.message);
                        if (!__classPrivateFieldGet(self, _Room_incomingMessageQueue, "f")[innerMessage.id]) {
                            __classPrivateFieldGet(self, _Room_incomingMessageQueue, "f")[innerMessage.id] = [];
                        }
                        __classPrivateFieldGet(self, _Room_incomingMessageQueue, "f")[innerMessage.id].push(innerMessage);
                        if (innerMessage.last) {
                            let wholeMessage = "";
                            for (let i = 0; i < __classPrivateFieldGet(self, _Room_incomingMessageQueue, "f")[innerMessage.id].length; i++) {
                                wholeMessage += __classPrivateFieldGet(self, _Room_incomingMessageQueue, "f")[innerMessage.id][i].payload;
                            }
                            delete __classPrivateFieldGet(self, _Room_incomingMessageQueue, "f")[innerMessage.id];
                            msg.message.message = wholeMessage;
                            self.notifier.notify(message.type, msg);
                        }
                    }
                    catch (e) {
                        this.logger.info("Failed to process inner message: " + msg.message);
                        self.notifier.notify(message.type, msg);
                    }
                }
            }
            else if (message.type === constants_1.SfuEvent.ACK && promises_1.default.promised(message.internalMessageId)) {
                promises_1.default.resolve(message.internalMessageId);
            }
            else {
                if (Object.values(constants_1.RoomEvent).includes(message.type)) {
                    self.notifier.notify(message.type, message);
                }
            }
        };
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            if (__classPrivateFieldGet(self, _Room__state, "f") === constants_1.RoomState.NEW) {
                __classPrivateFieldSet(self, _Room__state, constants_1.RoomState.JOINED, "f");
                try {
                    const offer = yield __classPrivateFieldGet(self, _Room__pc, "f").createOffer();
                    if (config) {
                        offer.sdp = __classPrivateFieldGet(self, _Room_instances, "m", _Room_applyContentTypeConfig).call(self, offer.sdp, config);
                    }
                    offer.sdp = offer.sdp.replace(/a=sendrecv/g, "a=sendonly");
                    yield __classPrivateFieldGet(self, _Room__pc, "f").setLocalDescription(offer);
                    const id = (0, uuid_1.v4)();
                    promises_1.default.add(id, resolve, reject);
                    self.connection.send(constants_1.InternalApi.JOIN_ROOM, {
                        id: self._id,
                        pin: self._pin,
                        sdp: offer.sdp,
                        nickname: nickname,
                        internalMessageId: id,
                        sdpType: constants_1.RemoteSdpType.OFFER
                    });
                }
                catch (e) {
                    reject(e);
                }
            }
            else {
                reject("Can't joined room with state " + __classPrivateFieldGet(this, _Room__state, "f"));
            }
        }));
    }
    ;
    updateState(config) {
        const self = this;
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            if (__classPrivateFieldGet(self, _Room__pc, "f").signalingState !== "stable") {
                reject("Peer connection signaling state is " + __classPrivateFieldGet(self, _Room__pc, "f").signalingState + ". Can't update room while negotiation is in progress");
                return;
            }
            try {
                const offer = yield __classPrivateFieldGet(self, _Room__pc, "f").createOffer();
                if (self._crutch.settingRemoteOffer) {
                    reject("Setting remote offer");
                    return;
                }
                offer.sdp = offer.sdp.replace(/a=sendrecv/g, "a=sendonly");
                yield __classPrivateFieldGet(self, _Room__pc, "f").setLocalDescription(offer);
                if (config) {
                    offer.sdp = __classPrivateFieldGet(self, _Room_instances, "m", _Room_applyContentTypeConfig).call(self, offer.sdp, config);
                }
                const id = (0, uuid_1.v4)();
                promises_1.default.add(id, resolve, reject);
                self.connection.send(constants_1.InternalApi.UPDATE_ROOM_STATE, {
                    id: self._id,
                    pin: self._pin,
                    sdp: offer.sdp,
                    internalMessageId: id,
                    sdpType: constants_1.RemoteSdpType.OFFER,
                    tid: this._crutch.tid
                });
            }
            catch (e) {
                reject(e);
            }
        }));
    }
    ;
    destroyRoom() {
        const self = this;
        return new Promise((resolve, reject) => {
            __classPrivateFieldSet(self, _Room__state, constants_1.RoomState.DISPOSED, "f");
            const id = (0, uuid_1.v4)();
            promises_1.default.add(id, resolve, reject);
            self.connection.send(constants_1.InternalApi.DESTROY_ROOM, {
                id: self._id,
                pin: self._pin,
                internalMessageId: id
            });
        });
    }
    ;
    leaveRoom() {
        const self = this;
        return new Promise((resolve, reject) => {
            __classPrivateFieldSet(self, _Room__state, constants_1.RoomState.DISPOSED, "f");
            const id = (0, uuid_1.v4)();
            promises_1.default.add(id, resolve, reject);
            self.connection.send(constants_1.InternalApi.LEAVE_ROOM, {
                id: self._id,
                pin: self._pin,
                internalMessageId: id
            });
            if (__classPrivateFieldGet(self, _Room__pc, "f")) {
                __classPrivateFieldGet(self, _Room__pc, "f").close();
                // zapp-28, react-native-webrtc fire 'connectionstatechange' byself
                if (typeof document !== 'undefined') {
                    __classPrivateFieldGet(self, _Room__pc, "f").dispatchEvent(new Event("connectionstatechange"));
                }
            }
        });
    }
    ;
    evictParticipant(nickname) {
        const self = this;
        return new Promise((resolve, reject) => {
            const id = (0, uuid_1.v4)();
            promises_1.default.add(id, resolve, reject);
            self.connection.send(constants_1.InternalApi.EVICT_PARTICIPANT, {
                id: self._id,
                nickname: nickname,
                internalMessageId: id
            });
        });
    }
    renameParticipant(nickname, newNickname) {
        const self = this;
        return new Promise(((resolve, reject) => {
            const id = (0, uuid_1.v4)();
            promises_1.default.add(id, resolve, reject);
            self.connection.send(constants_1.InternalApi.RENAME_PARTICIPANT, {
                roomId: self._id,
                nickname: nickname,
                newNickname: newNickname,
                internalMessageId: id
            });
        }));
    }
    sendMessage(msg) {
        const self = this;
        return new Promise((resolve, reject) => {
            const id = (0, uuid_1.v4)();
            /**
             * Note self in case of chunked message promise will be resolved only after sending last message (last: true)
             */
            promises_1.default.add(id, resolve, reject);
            const chunkSize = 16384;
            if (msg.length > chunkSize) {
                const chunks = msg.match(new RegExp("(.|[\r\n]){1," + chunkSize + "}", "g"));
                for (let i = 0; i < chunks.length; i++) {
                    __classPrivateFieldGet(self, _Room_instances, "m", _Room_dChannelSend).call(self, JSON.stringify({
                        id: id,
                        last: i === chunks.length - 1,
                        payload: chunks[i]
                    }));
                }
            }
            else {
                __classPrivateFieldGet(self, _Room_instances, "m", _Room_dChannelSend).call(self, JSON.stringify({
                    id: id,
                    payload: msg
                }));
            }
        });
    }
    ;
    changeQuality(trackId, quality, tid) {
        const self = this;
        return new Promise((resolve, reject) => {
            const id = (0, uuid_1.v4)();
            promises_1.default.add(id, resolve, reject);
            self.connection.send(constants_1.InternalApi.CHANGE_QUALITY, {
                roomId: self._id,
                id: trackId,
                quality: quality,
                tid: tid,
                internalMessageId: id
            });
        });
    }
    ;
    muteTrack(trackId, mute) {
        const self = this;
        return new Promise((resolve, reject) => {
            const id = (0, uuid_1.v4)();
            promises_1.default.add(id, resolve, reject);
            self.connection.send(constants_1.InternalApi.MUTE_TRACK, {
                roomId: self._id,
                id: trackId,
                mute: mute
            });
        });
    }
    ;
    on(event, callback) {
        this.notifier.add(event, callback);
        return this;
    }
    ;
    off(event, callback) {
        this.notifier.remove(event, callback);
        return this;
    }
    ;
    id() {
        return this._id;
    }
    name() {
        return this._name;
    }
    pin() {
        return this._pin;
    }
    nickname() {
        return this._nickname;
    }
    pc() {
        return __classPrivateFieldGet(this, _Room__pc, "f");
    }
    role() {
        return __classPrivateFieldGet(this, _Room__role, "f");
    }
    invite() {
        return __classPrivateFieldGet(this, _Room_inviteId, "f");
    }
    state() {
        return __classPrivateFieldGet(this, _Room__state, "f");
    }
    creationTime() {
        return __classPrivateFieldGet(this, _Room__creationTime, "f");
    }
}
exports.Room = Room;
_a = Room, _Room__state = new WeakMap(), _Room__role = new WeakMap(), _Room_inviteId = new WeakMap(), _Room__pc = new WeakMap(), _Room_incomingMessageQueue = new WeakMap(), _Room_dChannel = new WeakMap(), _Room__creationTime = new WeakMap(), _Room_instances = new WeakSet(), _Room_dChannelSend = function _Room_dChannelSend(msg) {
    this.logger.info("dchannel ", "==>", msg);
    __classPrivateFieldGet(this, _Room_dChannel, "f").send(msg);
}, _Room_applyContentTypeConfig = function _Room_applyContentTypeConfig(sdp, config) {
    let ret = "";
    for (const str of sdp.split("\n")) {
        if (str && str.length > 0) {
            ret += str + "\n";
            if (str.indexOf("a=msid:") > -1) {
                const msid = str.substring(str.indexOf(" ") + 1).trim();
                if (config[msid]) {
                    ret += constants_1.InternalApi.TRACK_CONTENT_HEADER + config[msid] + "\r\n";
                }
            }
        }
    }
    return ret;
}, _Room_setRemoteDescription = function _Room_setRemoteDescription(remoteSdp) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield __classPrivateFieldGet(this, _Room__pc, "f").setRemoteDescription(remoteSdp.info);
        }
        catch (e) {
            this.logger.warn(e);
            this.logger.debug("Was trying to set " + remoteSdp.info.sdp);
            this.logger.debug("Local sdp is " + __classPrivateFieldGet(this, _Room__pc, "f").localDescription.sdp);
            throw e;
        }
    });
};
_Room_CONTROL_CHANNEL = { value: "control" };
//# sourceMappingURL=room.js.map